from .rest_helper import RestPackageRegistry, RestPackage, RestPackageInstance
from af_org_model_manager.models.client_session_details import ClientSessionDetails
from typing import Type, TypeVar
from multiprocessing.pool import ThreadPool

T = TypeVar('T')

class ServiceProvider(object):
    """ Per user session service provider caching service stubs and providing authentication data to them.
    """
    __rest_package_registry: RestPackageRegistry = None
    # package to package instance map
    __rest_packages: [RestPackage, RestPackageInstance] = None
    # cached service stubs
    __services: [type, object] = None
    # authentication
    __csd: ClientSessionDetails = None
    # thread pool for async requests
    __async_thread_pool:ThreadPool = None

    def __init__(self, rest_package_registry: RestPackageRegistry, async_thread_pool:ThreadPool):
        self.__rest_package_registry = rest_package_registry
        self.__rest_packages = {}
        self.__services = {}
        self.__async_thread_pool = async_thread_pool
    
    
    def __get_package_instance(self, service_type:Type) -> RestPackageInstance:
        """ Returns the (cached) ApiClient instance for the package of the given service_type.
        Note, that service_type may also be a model class.
        """
        # find the package description and package instance
        pkg = self.__rest_package_registry.get_rest_package(service_type)
        pkg_instance: RestPackageInstance = None
        if pkg in self.__rest_packages:
            pkg_instance = self.__rest_packages[pkg]
        else:
            pkg_instance = RestPackageInstance(pkg, self.__async_thread_pool)

        return pkg_instance

    def get_service(self, service_type: Type[T]) -> T:
        """
        Returns a service instance for the given service type, e.g.
        get_service(InstanceControlApi)
        @param service_type The class of the requested service.
        """
        # return a cached instance
        if service_type in self.__services:
            return self.__services[service_type]

        # get the ApiClient object of the package
        api_client = self.__get_package_instance(service_type).api_client
        
        # authentication data
        if self.__csd:
            self.__use_authentication(api_client)
        # instantiate the service
        service = service_type(api_client)
        # cache it
        self.__services[service_type] = service
        return service

    def __use_authentication(self, api_client: object):
        """Uses the client session details to set the default header
        """
        api_client.set_default_header("x-AF-Security-Token", self.__csd.token)

    def authenticated(self, csd:ClientSessionDetails):
        """ set the authentication for all ApiClient objects
        """
        if self.__csd != None:
            raise Exception("Already authenticated")

        self.__csd = csd
        for _, inst in self.__rest_packages:
            self.__use_authentication(inst.api_client)

    def deserialize(self, data, klass):
        """ Deserialize data using the given class of the generated OpenAPI models. 
        """
        return self.__get_package_instance(klass).deserialize(data, klass)

    def serialize(self, obj):
        """ Serialize REST model object 
        """
        klass = type(obj)
        return self.__get_package_instance(klass).serialize(obj)



from . import AristaFlowClientService
from . import Configuration
from .rest_helper import RestPackageRegistry
from aristaflow.service_provider import ServiceProvider
from multiprocessing.pool import ThreadPool


class AristaFlowClientPlatform(object):
    """ Entry point to the AristaFlow Python Client framework.
    """

    # thread pool for async requests
    __async_thread_pool:ThreadPool = None

    def __init__(self, configuration: Configuration):
        self.configuration = configuration
        self.__client_services: [str, AristaFlowClientService] = {}
        self.__rest_package_registry = RestPackageRegistry(configuration)
        self.__async_thread_pool = ThreadPool(configuration.async_thread_pool_size)

    def get_client_service(self, user_session: str="python_default_session"):
        """
        :return: AristaFlowClientService The client service for the given user session
        """
        if user_session in self.__client_services:
            return self.__client_services[user_session]
        cs = AristaFlowClientService(
            self.configuration, user_session, ServiceProvider(self.__rest_package_registry, self.__async_thread_pool))
        self.__client_services[user_session] = cs
        return cs



# -*- coding: utf-8 -*-
from typing import List, Union

from af_execution_manager.api.instance_control_api import InstanceControlApi
from af_execution_manager.models.templ_ref_initial_remote_iterator_data import TemplRefInitialRemoteIteratorData
from af_execution_manager.api.templ_ref_remote_iterator_rest_api import TemplRefRemoteIteratorRestApi
from af_execution_manager.models.templ_ref_remote_iterator_data import TemplRefRemoteIteratorData
from af_execution_manager.models.template_reference import TemplateReference
from af_execution_manager.models.instance_creation_rest_data import InstanceCreationRestData
from aristaflow.utils import VERSION
from af_execution_manager.models.parameter_value import ParameterValue
from af_execution_manager.models.instance_creation_data import InstanceCreationData
from af_execution_manager.models.data_container import DataContainer
from aristaflow.abstract_service import AbstractService
class ProcessService(AbstractService):
    """ Process related methods
    """
    
    def __tpl_version_key(self, tpl:TemplateReference) -> int:
        return VERSION.key(tpl.version)
    
    def get_instantiable_template_by_type(self, process_type:str) -> TemplateReference:
        """ Finds the first instantiable template of the given process Type
        """
        tpls = self.get_instantiable_templates()
        tpls_of_type = []
        for tpl in tpls:
            if tpl.process_type == process_type:
                tpls_of_type += [tpl]
        # empty result
        
        if len(tpls_of_type) == 0:
            return None        
        # reverse sort by version
        tpls_of_type = sorted(tpls_of_type, key=self.__tpl_version_key, reverse=True)
        # return first entry
        return tpls_of_type[0]
    
    
    def get_instantiable_templates(self) -> List[TemplateReference]:
        """ Retrieves the instantiable tempaltes from the server
        :return: List[TemplateReference] The instantiable templates for the current user 
        """
        ic:InstanceControlApi = self._service_provider.get_service(InstanceControlApi)
        initial:TemplRefInitialRemoteIteratorData = ic.get_instantiable_templ_refs()
        tpls: List[TemplateReference] = []
        self.__iterate(tpls, initial)
        return tpls
    
    def __iterate(self, tpls: List[TemplateReference], inc: Union[TemplRefInitialRemoteIteratorData, TemplRefRemoteIteratorData]):
        """ Consumes an template reference remote iterator until it is used up
        @param tpls The tpls list to fill with the template references
        @param inc The first or next iteration to consume and append to tpls. 
        """
        if inc == None:
            return
        # append the tpls
        if inc.templ_refs:
            tpls += inc.templ_refs
        # iterator is used up
        if inc.dropped:
            return

        # fetch next
        tref_rest: TemplRefRemoteIteratorRestApi = self._service_provider.get_service(
            TemplRefRemoteIteratorRestApi)
        next_it: TemplRefRemoteIteratorData = tref_rest.templ_ref_get_next(inc.iterator_id)
        self.__iterate(tpls, next_it)

    def start_by_type(self, process_type:str, callback_uri:str = None, input_data:dict = None) -> str:
        """ Starts the newest version of the given process type, returns the logical ID of the started instance.
        """
        tpl = self.get_instantiable_template_by_type(process_type)
        if tpl == None:
            raise Exception('Unknown process type: ' + process_type)
        return self.start_by_id(tpl.id, callback_uri, input_data)

    def start_by_id(self, template_id:str, callback_uri:str = None, input_data:dict = None) -> str:
        """ Starts a process given by the template id. Returns the logical ID of the started instance.
        """
        ic:InstanceControlApi = self._service_provider.get_service(InstanceControlApi)
        if callback_uri == None:
            inst_creation_data = InstanceCreationData(sub_class='InstanceCreationData')
            inst_creation_data.dc = self.__create_instance_container(ic, template_id, input_data)
            return ic.create_and_start_instance(template_id, body=inst_creation_data)
        else:
            inst_creation_data = InstanceCreationRestData(sub_class='InstanceCreationRestData', notification_callback = callback_uri)
            inst_creation_data.dc = self.__create_instance_container(ic, template_id, input_data)
            return ic.create_and_start_instance_callback(body=inst_creation_data, templ_id=template_id)


    def __create_instance_container(self, ic:InstanceControlApi, template_id:str, input_data:dict) -> DataContainer:
        """ Creates an instance data container for the given template, if required
        """
        idc = None
        if input_data != None and len(input_data) != 0:
            idc = ic.create_instance_data_container(template_id)
            pv:ParameterValue = None
            for pv in idc.values:
                if pv.name in input_data:
                    value = input_data[pv.name]
                    pv.value = value
        return idc
        
        


# -*- coding: utf-8 -*-
import re
class Version(object):
    """ Parsing and comparing version strings
    """
    __pattern:re = re.compile("(\\d+)(?:\\.(\\d+))?(?:\\.(\\d+))?(?:\\.(\\d+))?(?:\\.(\\d+))?")
    
    def key(self, v:str) -> int:
        if not(v):
            return 0
        
        m:re = self.__pattern.match(v)
        # invalid version string
        if m == None:
            return 0
        
        gs = m.groups()
        key = 0;
        places = len(gs)
        for i in range(places):
            p = gs[i]
            if p == None:
                break
            if len(p) > 3:
                # ugh... unsupported
                return 0
            key += int(p) * 10**(places - i)

        #print(f"{v} = {key}")
        return key

VERSION = Version()

#print(v.compare("1.5.0", "1.5.0"))
#vs = ["999.999.999.999.999", "1.5.0", "2.3.555.0", "2.3.8", "5.1", "3", "0.9.6", "2.3.8.1", "2.3.5"]
#print(sorted(vs, key=v.key))



# -*- coding: utf-8 -*-
from aristaflow.service_provider import ServiceProvider
class AbstractService(object):
    ''' Abstract base class for service helpers
    '''

    _service_provider:ServiceProvider = None

    def __init__(self, service_provider:ServiceProvider):
        self._service_provider = service_provider



from typing import List, Union

from af_worklist_manager.api.inc_client_worklists_api import IncClientWorklistsApi
from af_worklist_manager.api.inc_worklist_update_api import IncWorklistUpdateApi
from af_worklist_manager.api.worklist_update_manager_api import WorklistUpdateManagerApi
from af_worklist_manager.models.client_worklist_item import ClientWorklistItem
from af_worklist_manager.models.client_worklist_item_update import ClientWorklistItemUpdate
from af_worklist_manager.models.inc_client_worklist_data import IncClientWorklistData
from af_worklist_manager.models.inc_worklist_update_data import IncWorklistUpdateData
from af_worklist_manager.models.initial_inc_client_worklist_data import InitialIncClientWorklistData
from af_worklist_manager.models.initial_inc_worklist_update_data import InitialIncWorklistUpdateData
from af_worklist_manager.models.update_interval import UpdateInterval
from af_worklist_manager.models.worklist_revision import WorklistRevision
from af_worklist_manager.models.worklist_update_configuration import WorklistUpdateConfiguration

from .service_provider import ServiceProvider
from .worklist_model import Worklist


class WorklistService(object):
    # The fetch size for incremental worklists / updates. If None, the
    # server-side default will be used
    fetch_count: int = None
    __worklist: Worklist = None
    __items: List[ClientWorklistItem] = None
    __service_provider:ServiceProvider = None

    def __init__(self, service_provider:ServiceProvider):
        self.__items = []
        self.__service_provider = service_provider

    def create_worklist_update_configuration(self) -> WorklistUpdateConfiguration:
        """ Creates a default worklist update configuration
        """
        update_intervals: list[UpdateInterval] = []
        # worklistFilter: NO_TL or TL_ONLY
        # notify_only: if True, the initial worklist will not be returned but
        # pushed
        wuc = WorklistUpdateConfiguration(
            update_mode_threshold=3000, update_intervals=update_intervals, worklist_filter="NO_TL", notify_only=False)
        return wuc

    def get_worklist(self) -> List[ClientWorklistItem]:
        """ Updates and returns the worklist of the current user
        """
        if self.__worklist != None:
            # perform update
            return self.update_worklist()

        wum: WorklistUpdateManagerApi = self.__service_provider.get_service(
            WorklistUpdateManagerApi)
        update_conf: WorklistUpdateConfiguration = self.create_worklist_update_configuration()
        wlit: InitialIncClientWorklistData = None
        if self.fetch_count != None:
            wlit = wum.logon_and_create_client_worklist(
                body=update_conf, count=self.fetch_count)
        else:
            wlit = wum.logon_and_create_client_worklist(body=update_conf)

        # currently no items in the worklist
        if wlit == None:
            return self.__items

        self.__iterate(self.__items, wlit)

        # remember the current worklist meta data
        self.__worklist = Worklist(
            wlit.worklist_id, wlit.revision, wlit.client_worklist_id, update_conf)

        return self.__items

    def __iterate(self, items: List[ClientWorklistItem], inc: Union[InitialIncClientWorklistData, IncClientWorklistData]):
        """ Consumes an incremental client worklist until its iterator is used up
        @param items The items list to fill with the update(s)
        @param inc The first or next iteration to consume and append to items. 
        """
        if inc == None:
            return
        # append the items
        if inc.items_flat:
            items += inc.items_flat
        # iterator is used up
        if inc.dropped:
            return

        # fetch next
        inc_cl: IncClientWorklistsApi = self.__service_provider.get_service(
            IncClientWorklistsApi)
        next_it: IncClientWorklistData = inc_cl.inc_client_wl_get_next(
            inc.inc_wl_id)
        self.__iterate(items, next_it)

    def update_worklist(self) -> List[ClientWorklistItem]:
        """ Updates the user's worklist and returns the items
        """
        if self.__worklist == None:
            return self.get_worklist()

        wu: WorklistUpdateManagerApi = self.__service_provider.get_service(
            WorklistUpdateManagerApi)
        inc_updts: InitialIncWorklistUpdateData = wu.get_worklist_updates(self.__worklist.worklist_id, body=self.__worklist.revision,
                                                                          filter=self.__worklist.wu_conf.worklist_filter)

        if inc_updts != None:
            updates: List[ClientWorklistItemUpdate] = []
            self.__iterate_updates(updates, inc_updts)
            self.__apply_worklist_updates(
                inc_updts.source_revision, inc_updts.target_revision, updates)
        return self.__items

    def __iterate_updates(self, updates: List[ClientWorklistItemUpdate], inc: IncWorklistUpdateData):
        """ Consumes the given worklist update iterator and appends all retrieved updates to the provided updates list. 
        """
        if inc == None:
            return
        if inc.item_updates:
            updates += inc.item_updates

        if inc.dropped:
            return

        # fetch next
        iwua: IncWorklistUpdateApi = self.__service_provider.get_service(
            IncWorklistUpdateApi)
        next_it: IncWorklistUpdateData = iwua.inc_wl_updt_get_next(
            inc.inc_upd_id)
        self.__iterate_updates(updates, next_it)

    def __apply_worklist_updates(self, source_revision: WorklistRevision, target_revision: int,
                                 updates: List[ClientWorklistItemUpdate]):
        """ Applies the provided worklist updates to self.__items. Checks the consistency to the source revision,
            and performs a full update if the state does not fit. Sets the new target revision in self.__worklist.
        """
        if self.__worklist.revision.update_count != source_revision.update_count or self.__worklist.revision.initialisation_date != source_revision.initialisation_date:
            # out of order update, clear and re-fetch everything
            self.__items.clear()
            self.__worklist = None
            self.get_worklist()
            return

        #print(f'Applying {len(updates)} updates')
        #print(updates)

        for update in updates:
            self.__apply_worklist_update(update)

        # remember the update count for the next update
        self.__worklist.revision.update_count = target_revision

    def __apply_worklist_update(self, update: ClientWorklistItemUpdate):
        """ Applies the given update to __items
        """
        update_type = update.update_type
        item = update.item
        if update_type == "CHANGED":
            self.__replace_or_add_item(item)
        elif update_type == "ADDED":
            self.__items += [item]
        elif update_type == "REMOVED":
            self.__remove_item(item)
        elif update_type == "ADDED_OR_CHANGED":
            self.__replace_or_add_item(item)
        elif update_type == "REMOVED_OR_NOTHING":
            self.__remove_item(item)
        else:
            raise RuntimeError("Unknown update type: " + update_type)

    def __replace_or_add_item(self, item: ClientWorklistItem):
        """ Replaces or adds the given worklist item in self.__items
        """
        #print('__replace_or_add_item: __items=', self.__items)
        for i in range(len(self.__items)):
            val = self.__items[i]
            if item.id == val.id:
                self.__items[i] = item
                return
        # not found above, append it
        self.__items.append(item)

    def __remove_item(self, item: ClientWorklistItem):
        """ Removes the given worklist item from self.__items
        """
        for val in self.__items:
            if item.id == val.id:
                self.__items.remove(val)
                return

    def find_item_by_id(self, item_id:str) -> ClientWorklistItem:
        """ Finds a worklist item by its worklist item id. Returns none, if not in the worklist of the user.
        """
        #print(f'Finding item with id {item_id}')
        self.update_worklist()
        #print(self.__items)
        for item in self.__items:
            if item.id == item_id:
                #print('Found')
                return item
        return None


'''
Worklist model classes
'''
from af_worklist_manager.models.worklist_revision import WorklistRevision
from af_worklist_manager.models.worklist_update_configuration import WorklistUpdateConfiguration


class Worklist(object):
    ''' Simply aggregates the few relevant properties of a worklist
    '''

    worklist_id: str = None
    revision: WorklistRevision = None
    client_worklist_id: int = None
    wu_conf: WorklistUpdateConfiguration = None

    def __init__(self, worklist_id: str, revision: WorklistRevision, client_worklist_id: int, wu_conf: WorklistUpdateConfiguration):
        '''
        Constructor
        '''
        self.worklist_id = worklist_id
        self.revision = revision
        self.client_worklist_id = client_worklist_id
        self.wu_conf = wu_conf
        pass



from typing import TypeVar, Type

from af_org_model_manager.api.global_security_manager_api import GlobalSecurityManagerApi
from af_org_model_manager.models.client_session_details import ClientSessionDetails

from .configuration import Configuration
#from .rest_helper import RestPackageRegistry, RestPackage, RestPackageInstance
from .worklist_service import WorklistService
from aristaflow.service_provider import ServiceProvider
from af_org_model_manager.models.authentication_data import AuthenticationData
from af_org_model_manager.models.auth_data_org_pos import AuthDataOrgPos
from af_org_model_manager.models.auth_data_user_name import AuthDataUserName
from af_org_model_manager.models.auth_data_arbitrary import AuthDataArbitrary
import base64
from af_org_model_manager.models.auth_data_app_name import AuthDataAppName
from af_org_model_manager.models.auth_data_password import AuthDataPassword
from af_org_model_manager.models.qualified_agent import QualifiedAgent
from af_worklist_manager.models.worklist_item import WorklistItem
from af_remote_html_runtime_manager.api.runtime_manager_api import RuntimeManagerApi
from af_remote_html_runtime_manager.api.synchronous_activity_starting_api import SynchronousActivityStartingApi
from af_remote_html_runtime_manager.models.gui_context import GuiContext
from af_remote_html_runtime_manager.models.ebp_instance_reference import EbpInstanceReference
from af_worklist_manager.models.af_activity_reference import AfActivityReference
from aristaflow.html_gui_context import HtmlGuiContext
from af_execution_manager.api.activity_execution_control_api import ActivityExecutionControlApi
import af_execution_manager
from af_remote_html_runtime_manager.models.activity_rest_callback_data import ActivityRestCallbackData
from aristaflow.process_service import ProcessService
from aristaflow.delegation_service import DelegationService


T = TypeVar('T')
D = TypeVar('D')


class AristaFlowClientService(object):
    """Client Services for accessing the AristaFlow BPM platform
    """
    # the user session id this client service belongs to
    __user_session: str = None
    # authentication
    __csd: ClientSessionDetails = None
    # the AristaFlow configuration
    __af_conf: Configuration = None
    __service_provider:ServiceProvider = None
    __worklist_service:WorklistService = None
    __is_html_runtime_manager_logged_on = False
    __process_service:ProcessService = None
    __delegation_service:DelegationService = None

    def __init__(self, configuration: Configuration,
                 user_session: str,
                 service_provider:ServiceProvider):
        self.__af_conf = configuration
        self.__user_session = user_session
        self.__service_provider = service_provider

    def get_service(self, service_type: Type[T]) -> T:
        """
        Returns a service instance for the given service type, e.g.
        get_service(InstanceControlApi)
        @param service_type The class of the requested service.
        """
        return self.__service_provider.get_service(service_type)

    @property
    def client_session_details(self) -> ClientSessionDetails:
        """ Returns the client session details of this service
        :return: The client session details of this service
        """
        return self.__csd

    @property
    def is_authenticated(self) -> bool:
        """ Returns true, if this client service is already authenticated
        """
        return self.__csd != None

    def authenticate(self, username: str, password: str = None, org_pos_id: int=None):
        if self.__csd != None:
            raise Exception("Already authenticated")
        
        auth_data:list[AuthenticationData] = []
        auth_data.append(AuthDataUserName(username, sub_class="AuthDataUserName"))
        if org_pos_id != None:
            auth_data.append(AuthDataOrgPos(org_pos_id, sub_class="AuthDataOrgPos"))
        psk = self.__af_conf.pre_shared_key
        method:str = None
        # if a password was provided, use it
        if password:
            auth_data.append(AuthDataPassword(password=password, sub_class="AuthDataPassword"))
            method = 'UTF-8_PASSWORD';
        # if PSK is configured, use that
        elif psk:
            # get the utf-8 bytes, encode them using base 64 and decode the resulting bytes using ASCII
            psk_encoded = base64.b64encode(bytes(psk, "UTF-8")).decode('ascii')
            auth_data.append(AuthDataArbitrary(data=psk_encoded, sub_class="AuthDataArbitrary"))
            method = 'SHARED_UTF-8_KEY'
        else:
            raise Exception('No authentication method left');
        
        gsm: GlobalSecurityManagerApi = self.get_service(
            GlobalSecurityManagerApi)

        # use a provided application name
        if self.__af_conf.application_name:
            if  org_pos_id == None:
                # if an application name is provided, an org position ID must be used as well
                # get the org positions
                agents:list[QualifiedAgent] = gsm.pre_authenticate_method(method, body=auth_data)
                agent:QualifiedAgent = None
                # pick the single org position
                if len(agents) == 1:
                    agent = agents[0]
                # none: can't login
                elif len(agents) == 0:
                    raise Exception(f"User does not have an org position {username} (supplied org position id: {org_pos_id})")
                else:
                    # use the first org position, except there is a agent_name/username match     
                    agent = agents[0]
                    for a in agents:
                        if a.agent_name == username:
                            agent = a
                            break
                # set the org position for the actual authentication
                auth_data.append(AuthDataOrgPos(agent.org_pos_id, sub_class='AuthDataOrgPos'))
            # use the application name
            auth_data.append(AuthDataAppName(app_name=self.__af_conf.application_name, sub_class="AuthDataAppName"))
        
        csds: list[ClientSessionDetails] = gsm.authenticate_all_method(method, self.__af_conf.caller_uri,
                                                                       body=auth_data)

        csd: ClientSessionDetails = None
        if len(csds) == 1:
            csd = csds[0]
        elif len(csds) == 0:
            raise Exception("User does not have an org position " +
                            username + " (supplied org position id: " + org_pos_id + ")")
        else:
            # pick the first as default
            csd = csds[0]
            # pick the one where username and org position name are the same
            for entry in csds:
                if entry.agent.agent.org_pos_name == entry.agent.agent.agent_name:
                    csd = entry
                    break

        self.__service_provider.authenticated(csd)
        self.__csd = csd

    def is_html_activity(self, item:WorklistItem) -> bool:
        return item.act_ref.gui_context_id == 'HTMLContext' or item.act_ref.executable_component_name == 'de.aristaflow.form.Form' or item.act_ref.executable_component_name  == 'de.aristaflow.form.GeneratedForm'

    @property
    def worklist_service(self):
        if self.__worklist_service == None:
            self.__worklist_service = WorklistService(self.__service_provider)
        return self.__worklist_service

    @property
    def process_service(self):
        if self.__process_service == None:
            self.__process_service = ProcessService(self.__service_provider)
        return self.__process_service

    @property
    def delegation_service(self):
        if self.__delegation_service == None:
            self.__delegation_service = DelegationService(self.__service_provider)
        return self.__delegation_service

    def start_html_activity(self, item:WorklistItem, callback_url:str):
        """
        Starts the given HTML GUI worklist item using the Remote HTML Runtime Manager
        """
        if item == None:
            raise Exception('No worklist item provided')
        # accept user form and HTMLContext based activities
        if not(self.is_html_activity(item)):
            raise Exception(f'Not an HTML activity: {item.act_ref.gui_context_id}')
        if item.state == 'STARTED':
            raise Exception('Item is already started')
        sas = self.get_html_activity_starting()
        gc:GuiContext = None
        ar:AfActivityReference = item.act_ref
        #print('Starting activity...')
        ebp_ir:EbpInstanceReference = EbpInstanceReference(ar.type, ar.instance_id, ar.instance_log_id, ar.base_template_id, ar.node_id, ar.node_iteration, ar.execution_manager_uris, ar.runtime_manager_uris)
        # "AVAILABLE", "ASSIGNED", "STARTED", "SUSPENDED", "ENQUIRED"
        if item.state == 'AVAILABLE' or item.state == 'ASSIGNED':
            if callback_url != None:
                cb_data = ActivityRestCallbackData(sub_class='ActivityRestCallbackData', notification_callback=callback_url, activity=ebp_ir)
                gc = sas.start_activity_callback(body=cb_data)
            else:
                gc = sas.start_activity(body=ebp_ir)
        else:
            if callback_url != None:
                cb_data = ActivityRestCallbackData(sub_class='ActivityRestCallbackData', notification_callback=callback_url, activity=ebp_ir)
                gc = sas.resume_activity_callback(body=cb_data)
            else:
                gc = sas.resume_activity(body=ebp_ir)
        return HtmlGuiContext(gc)
    
    
    def get_html_activity_starting(self) -> SynchronousActivityStartingApi:
        """
        Returns the Remote HTML Runtime Manager Syncrounous Activity Starting, ensuring logon to the Runtime Manager
        """
        sas:SynchronousActivityStartingApi = self.get_service(SynchronousActivityStartingApi)
        if self.__is_html_runtime_manager_logged_on:
            return sas
        rm:RuntimeManagerApi = self.get_service(RuntimeManagerApi)
        rm.logon(body=self.__csd)
        self.__is_html_runtime_manager_logged_on = True
        return sas
    
    def reset_activity(self, item:WorklistItem):
        """ Resets the given worklist item.
        """
        if item.state != 'STARTED':
            # nothing to do
            return
        # TODO check for a "local" activity for a soft reset
        ar:AfActivityReference = item.act_ref
        ebp_ir = af_execution_manager.EbpInstanceReference(ar.type, ar.instance_id, ar.instance_log_id, ar.base_template_id, ar.node_id, ar.node_iteration, ar.execution_manager_uris, ar.runtime_manager_uris)
        aec:ActivityExecutionControlApi = self.get_service(ActivityExecutionControlApi)
        aec.reset_to_prev_savepoint(body=ebp_ir, force=True)
        
    def deserialize(self, data, klass:Type[D]) -> D:
        """ Deserialize data using the given class of the generated OpenAPI models. 
        """
        return self.__service_provider.deserialize(data, klass)

    def serialize(self, obj) -> str:
        """ Serialize REST model object 
        """
        return self.__service_provider.serialize(obj)
    
    @property
    def autostart_timeout_seconds(self) -> int:
        """ Wait time in seconds for auto start signals
        """
        return self.__af_conf.autostart_timeout_seconds



# -*- coding: utf-8 -*-

from aristaflow.configuration import Configuration
from aristaflow.client_service import AristaFlowClientService
from aristaflow.client_platform import AristaFlowClientPlatform
from aristaflow.worklist_service import WorklistService



# coding: utf-8

"""

"""

from setuptools import setup, find_packages  # noqa: H301

# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = ["urllib3 >= 1.15", "six >= 1.10", "certifi", "python-dateutil"]

setup(
    name="aristaflowpy",
    version="1.5.0",
    description="AristaFlow BPM",
    author_email="info@aristaflow.com",
    url="https://www.aristaflow.com",
    keywords=["AristaFlow", "BPM", "Workflow", "REST"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description="""\
        AristaFlow BPM Python integration
        """
)



# -*- coding: utf-8 -*-
from aristaflow.abstract_service import AbstractService
from af_worklist_manager.models.worklist_item import WorklistItem
from af_worklist_manager.api.delegation_manager_api import DelegationManagerApi
from af_worklist_manager.api.del_rec_remote_iterator_api import DelRecRemoteIteratorApi
from af_worklist_manager.models.qa_initial_remote_iterator_data import QaInitialRemoteIteratorData
from typing import List, Union, Dict
from af_worklist_manager.models.qualified_agent import QualifiedAgent
from af_worklist_manager.models.qa_remote_iterator_data import QaRemoteIteratorData
from af_worklist_manager.models.delg_rec_with_comment import DelgRecWithComment
class DelegationService(AbstractService):
    '''
    Helper methods for implementing delegation
    '''

    def get_delegation_recipients(self, *items:WorklistItem) -> List[QualifiedAgent]:
        """ Retrieves the possible delegation recipients for the given worklist items, ie.
            the intersection of all all possible delegation recipients of all items.
        """
        recipient_set:Dict[str, QualifiedAgent] = None
        for item in items:
            recs = self._get_delegation_recipients(item)
            rec:QualifiedAgent = None
            if recipient_set == None:
                # initial iteration: add what is there
                recipient_set = {}
                for rec in recs:
                    recipient_set[f'{rec.agent_id}_{rec.org_pos_id}'] = rec
            else:
                # not the initial iteration: remove what doesn't exist
                keys_to_delete:List[str] = []
                for added_rec_key in recipient_set:
                    added_rec = recipient_set[added_rec_key]
                    found = False
                    for rec in recs:
                        if rec.agent_id == added_rec.agent_id and rec.org_pos_id == added_rec.org_pos_id:
                            found = True
                            break
                    if found == False:
                        keys_to_delete.append(added_rec_key)
                # remove from target set all entries which are not found in the delegation
                # recipients of the current item_id
                for key_to_delete in keys_to_delete:
                    del recipient_set[key_to_delete]

            # empty intersection, no need to continue
            if len(recipient_set) == 0:
                break

        return recipient_set.values()

    def _get_delegation_recipients(self, item:WorklistItem) -> List[QualifiedAgent]:
        """ Retrieves the possible delegation recipients for the given worklist item
        """
        dm:DelegationManagerApi = self._service_provider.get_service(DelegationManagerApi)
        drec:QaInitialRemoteIteratorData = dm.get_delegation_recipients(item.id)
        recipients:List[QualifiedAgent] = []
        self._iterate_delegation_recipients(recipients, drec)
        return recipients
    
    
    def _iterate_delegation_recipients(self, recipients:List[QualifiedAgent], cur_iter:Union[QaInitialRemoteIteratorData, QaRemoteIteratorData]):
        """ Reads up the given delegation recipients iterator
        """
        if cur_iter == None:
            return 
        if cur_iter.agents:
            recipients += cur_iter.agents
        if cur_iter.dropped:
            return
        
        api:DelRecRemoteIteratorApi = self._service_provider.get_service(DelRecRemoteIteratorApi)
        next_iter = api.del_rec_get_next(cur_iter.iterator_id)
        self._iterate_delegation_recipients(recipients, next_iter)

    def delegate(self, item:WorklistItem, comment:str, *args):
        """ Delegates the given worklist item using the given comments to all QualifiedAgents supplied in *args
            :param List[QualifiedAgents] args: Delegation recipients
        """
        dm:DelegationManagerApi = self._service_provider.get_service(DelegationManagerApi)
        body:DelgRecWithComment = DelgRecWithComment(recipients=args, comment=comment)
        dm.delegate_work_item(body, item.id)
        
        
        


from af_remote_html_runtime_manager.models.gui_context import GuiContext

class HtmlGuiContext(object):
    '''
    Small wrapper given direct access to Remote HTML GUI Context attributes.
    '''
    _gui_context:GuiContext = None

    def __init__(self, gui_context:GuiContext):
        self._gui_context = gui_context
        
    @property
    def gui_context(self) -> GuiContext:
        return self._gui_context
    
    @property
    def url(self) -> str:
        return self._gui_context.gui_context_attributes['URLContext.URL']



from importlib import import_module

from aristaflow.configuration import Configuration
from multiprocessing.pool import ThreadPool


class RestPackage(object):
    """
    Represents an AristaFlow REST endpoint package, e.g. af_execution_manager.
    """
    package_name: str = None
    # cached configuration as this can be used across all instances
    __config: object = None
    af_conf: Configuration = None

    def __init__(self, package_name: str, af_conf: Configuration):
        """
        :param package_name str: The base package name, e.g. af_org_model_manager
        :param af_conf Configuration: The AristaFlow BPM configuration.
        """
        self.package_name = package_name
        self.af_conf = af_conf

    @property
    def config(self):
        """
        :return: The configuration of the REST package
        """
        if self.__config == None:
            config_module = import_module(self.package_name + ".configuration")
            config_class = config_module.__getattribute__("Configuration")
            config = config_class()
            config.host = self.af_conf.get_host(self.service_type_name)
            config.debug = self.af_conf.get_debug(self.service_type_name)
            config.verify_ssl = self.af_conf.verify_ssl
            self.__config = config

        return self.__config

    def build_api_client(self):
        """
        Create a new ApiClient instance for this REST package
        """
        ac_module = import_module(self.package_name + ".api_client")
        ac_class = ac_module.__getattribute__("ApiClient")
        api_client = ac_class(self.config)
        api_client.set_default_header(
            "x-AF-Caller-URI", self.af_conf.caller_uri)
        return api_client

    @property
    def service_type_name(self):
        """
        The AristaFlow service type name for this package, e.g. ExecutionManager
        """
        if self.package_name == "af_execution_manager":
            return "ExecutionManager"
        elif self.package_name == "af_licence_manager":
            return "LicenceManager"
        elif self.package_name == "af_org_model_manager":
            return "OrgModelManager"
        elif self.package_name == "af_process_manager":
            return "ProcessManager"
        elif self.package_name == "af_runtime_service":
            return "RuntimeService"
        elif self.package_name == "af_worklist_manager":
            return "WorklistManager"
        elif self.package_name == "af_remote_html_runtime_manager":
            return "RuntimeManager"
        elif self.package_name == "af_simple_process_image_renderer":
            return "SimpleProcessImageRenderer"
        elif self.package_name == "af_process_image_renderer":
            return "ProcessImageRenderer"
        else:
            raise Exception("Unknown package: " + self.package_name)

    @property
    def default_instance_name(self):
        """
        The AristaFlow default hierarchical instance name, e.g. /ExecutionManager/ExecutionManager
        """
        service_instance_name: str = self.service_type_name if self.service_type_name != "RuntimeManager" else "RemoteHTMLRuntimeManager"
        return "/" + self.service_type_name + "/" + service_instance_name


class RestPackageRegistry(object):
    __rest_packages: [str, RestPackage] = None
    af_conf: Configuration = None

    def __init__(self, af_conf: Configuration):
        self.__rest_packages = {}
        self.af_conf = af_conf

    def get_rest_package(self, service_type: type):
        """
        Returns the REST package for the given service type.
        :return RestPackage: The REST package object.
        """
        pn = self.get_package_name(service_type)
        if pn in self.__rest_packages:
            return self.__rest_packages[pn]
        pkg = RestPackage(pn, self.af_conf)
        self.__rest_packages[pn] = pkg
        return pkg

    def get_package_name(self, service_type: type):
        """
        Returns the rest package for the given service type.
        :return str: The package name for the given service type.
        """
        moduleName: str = service_type.__module__
        return moduleName.split(".")[0]


class RestPackageInstance(object):
    """
    A REST package instance, ie. a REST package with an ApiClient object to be used for
    all of its services
    """
    __rest_package: RestPackage = None
    __api_client: object = None

    def __init__(self, rest_package: RestPackage, async_thread_pool:ThreadPool):
        self.__rest_package = rest_package
        self.__async_thread_pool = async_thread_pool

    @property
    def api_client(self) -> object:
        if self.__api_client == None:
            self.__api_client = self.__rest_package.build_api_client()
            # replace the default thread pool with our own, shared one
            orig_pool = self.__api_client.pool
            self.__api_client.pool = self.__async_thread_pool
            orig_pool.close() # no need to join(), work hadn't started yet
        return self.__api_client
    
    def deserialize(self, data, klass):
        """ Deserialize data using the given class of the generated OpenAPI models. 
        """
        ac = self.api_client
        #print(dir(ac))
        return ac._ApiClient__deserialize(data, klass)
    
    def serialize(self, obj):
        """ Serialize REST model object 
        """
        ac = self.api_client
        return ac.sanitize_for_serialization(obj)
